public import "ecere"

/*
   Icosahedron Snyder equal-area (ISEA) and ISEA 5x6 2DTMS projections
   --------------------------------------------------------------------------
   This module includes code adapted from Java to eC from
   Franz-Benjamin Mocnik's ISEA implementation found at
   https://github.com/mocnik-science/geogrid/blob/master/
    src/main/java/org/giscience/utils/geogrid/projections/ISEAProjection.java
   with the following license:

   MIT License

   Copyright (c) 2017-2019 Heidelberg University

   Permission is hereby granted, free of charge, to any person obtaining a copy
   of this software and associated documentation files (the "Software"), to deal
   in the Software without restriction, including without limitation the rights
   to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
   copies of the Software, and to permit persons to whom the Software is
   furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in all
   copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
   OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
   SOFTWARE.
*/

/* The ISEA projection a projects a sphere on the icosahedron. Thereby the size of areas mapped to the icosahedron
 * are preserved. Angles and distances are however slightly distorted. The angular distortion is below 17.27 degree, and
 * the scale variation is less than 16.3 per cent.
 *
 * The projection has been proposed and has been described in detail by:
 *
 * John P. Snyder: An equal-area map projection for polyhedral globes. Cartographica, 29(1), 10–21, 1992.
 * doi:10.3138/27H7-8K88-4882-1752
 *
 * Another description and improvements can be found in:
 *
 * Erika Harrison, Ali Mahdavi-Amiri, and Faramarz Samavati: Optimization of inverse Snyder polyhedral projection.
 * International Conference on Cyberworlds 2011. doi:10.1109/CW.2011.36
 *
 * Erika Harrison, Ali Mahdavi-Amiri, and Faramarz Samavati: Analysis of inverse Snyder optimizations.
 * In: Marina L. Gavrilova, and C. J. Kenneth Tan (Eds): Transactions on Computational Science XVI. Heidelberg,
 * Springer, 2012. pp. 134–148. doi:10.1007/978-3-642-32663-9_8
 */

public class GeoProjection
{
   virtual bool geoToCartesian(const GeoPoint3D position, Vector3D result, int lonQuad);
   virtual bool cartesianToGeo(const Vector3D position, GeoPoint3D result);
};
public struct GeoPoint { Degrees lat, lon; };
public struct GeoPoint3D { Degrees lat, lon, alt; };
public struct GeoExtent { GeoPoint ll, ur; };

static define wgs84InvFlattening = 298.257223563;
static define wgs84Major = 6378137.0;  // Meters
static define wgs84Minor = wgs84Major - (wgs84Major / wgs84InvFlattening); // 6356752.3142451792955399
static define wholeWorld = GeoExtent { { -90, -180 }, { 90, 180 } };

static define phi = (1 + sqrt(5)) / 2;
// radius
static define a2 = wgs84Major * wgs84Major;
static define c2 = wgs84Minor * wgs84Minor;

static define eccentricity = 0.0818191908426; // sqrt(1.0 - (c2/a2));
static define log1pe_1me = 0.1640050079196;   // log((1+e)/(1-e)))
static define S = Pi * (2 * a2 + c2/eccentricity * log1pe_1me);
static define earthAuthalicRadius = 6371007.18091875; //sqrt(S / (4 * Pi)); // R -- authalic sphere radius for WGS84 [6371007.1809184747 m]
static define R2 = earthAuthalicRadius * earthAuthalicRadius; // R^2
static define RprimeOverR = 0.9103832815095032; // (1 / (2 * sqrt(5)) + 1 / 6.0) * sqrt(Pi * sqrt(3)); // R' / R
static define Rprime = RprimeOverR * earthAuthalicRadius; // R'

// distortion
// static double maximumAngularDistortion = 17.27;
// static double maximumScaleVariation = 1.163;
// static double minimumScaleVariation = .860;

// Vertices of dodecahedron centered in icosahedron faces
static define E = Degrees { 52.6226318593487 }; //(Degrees)atan((3 + sqrt(5)) / 4); // Latitude of center of top icosahedron faces
static define F = Degrees { 10.8123169635739 }; //(Degrees)atan((3 - sqrt(5)) / 4); // Latitude of center of faces mirroring top icosahedron face
static define numIcosahedronFaces = 20;
static const GeoPoint facesCenterDodecahedronVertices[numIcosahedronFaces] =
{
   {  E, -144 }, {  E, -72 }, {  E,  0 }, {  E,  72 }, {  E, 144 },
   {  F, -144 }, {  F, -72 }, {  F,  0 }, {  F,  72 }, {  F, 144 },
   { -F, -108 }, { -F, -36 }, { -F, 36 }, { -F, 108 }, { -F, 180 },
   { -E, -108 }, { -E, -36 }, { -E, 36 }, { -E, 108 }, { -E, 180 }
};
// static define precision = Degrees { 1e-9 };
static define precision = Degrees { 1e-11 };
static define precisionPerDefinition = Degrees { 1e-5 };

static define Rprime2X = 2 * Rprime;
static define AzMax = Degrees { 120 };
static define sdc2vos = F + 2 * Degrees { 58.2825255885418 }/*(Degrees)atan(phi)*/ - Degrees { 90 }; // g -- sphericalDistanceFromCenterToVerticesOnSphere
static define tang = 0.763932022500419; //tan(sdc2vos);
static define cotTheta = 1.0 / tan(Degrees { 30 });
static define RprimeTang = Rprime * tang; // twice the center-to-base distance
static define centerToBase = RprimeTang / 2;
static define triWidth = RprimeTang * 1.73205080756887729352744634150587236694280525381038; //sqrt(3);
static define Rprime2Tan2g = RprimeTang * RprimeTang;
static define cosG = cos(Degrees { 36 });
static define sinGcosSDC2VoS = sin(Degrees { 36 }) * cos(sdc2vos); // sin G cos g
static define westVertexLon = Degrees { -144 };
static const double yOffsets[4] = { -2*centerToBase, -4*centerToBase, -5*centerToBase, -7*centerToBase };

__attribute__((unused)) static inline Radians latGeodeticToParametric(Degrees phi)
{
   static const double boa = wgs84Minor / wgs84Major;
   return atan(tan(phi) * boa);
}

__attribute__((unused)) static inline Radians latParametricToGeodetic(Degrees beta)
{
   static const double aob = wgs84Major / wgs84Minor;
   return atan(tan(beta) * aob);
}

static inline Radians latGeodeticToGeocentric(Radians phi)
{
   static const double b2oa2 = (wgs84Minor * wgs84Minor) / (wgs84Major * wgs84Major);
   return atan(tan(phi) * b2oa2);
}

static inline Radians latGeocentricToGeodetic(Radians theta)
{
   static const double a2ob2 = (wgs84Major * wgs84Major) / (wgs84Minor * wgs84Minor);
   return atan(tan(theta) * a2ob2);
}

public struct ISEAFacePoint
{
   int face;
   double x, y;
};

// #define POLAR_ORIENTATION

public class ISEAPlanarProjection : GeoProjection
{
public:
   property GeoPoint orientation
   {
      set { orientation = value; sinOrientationLat = sin(value.lat); cosOrientationLat = cos(value.lat); }
      get { value = orientation; }
   }

#ifdef POLAR_ORIENTATION
   // Polar orientation (+proj=isea +orient=pole)
   property::orientation = { 0, 0 };

   /*
    * One corner of the icosahedron is, by default, facing to the north pole, and one to the south pole. The provided
    * orientation is relative to the default orientation.
    *
    * The orientation shifts every location by the angle orientation.lon in direction of positive
    * longitude, and thereafter by the angle orientation.lat in direction of positive latitude.
    */
#else
   // Orientation symmetric to equator (+proj=isea)
   property::orientation = { (E + F) / 2 /* 90 - 58.282525590 = 31.7174744114613 */, -11.25 };

   /* Sets the orientation of the icosahedron such that the north and the south poles are mapped to the edge midpoints
    * of the icosahedron. The equator is thus mapped symmetrically.
    */
#endif

   bool geoToCartesian(const GeoPoint3D position, Vector3D result, int lonQuad)
   {
      ISEAFacePoint p { face = 0 };
      Radians lat = latGeodeticToGeocentric(position.lat);

      if(sphereToIcosahedron({ lat, position.lon }, p))
      {
         if(p.face >= 0 && p.face <= 19)
         {
            int fy = p.face / 5, fx = p.face - 5*fy;
            double x = p.x + (2*fx + fy/2 + 1) * triWidth/2;
            double y = p.y + yOffsets[fy] + 3*centerToBase;
            result = { x = x, y = y, z = p.face + 1 }; // Storing face in z for now
         }
      }
      return p.face != -1;
   }

   bool cartesianToGeo(const Vector3D position, GeoPoint3D result)
   {
      bool r = false;
      static const double epsilon = 1E-11; //0.000000001;
      int face = (int)position.z;
      if(!face)
      {
         // Rotate and shear to determine face if not stored in position.z
         static const double sr = -0.86602540378443864676372317075293618347140262690519, cr = 0.5;    // sin(-60), cos(-60)
         static const double shearX = 1.0 / 1.73205080756887729352744634150587236694280525381038; //sqrt(3); // 0.5773502691896257 -- 2*centerToBase / triWidth
         static const double sx = 1.0 / triWidth, sy = 1.0 / (3*centerToBase);
         double yp = -(position.x * sr + position.y * cr);
         double x = (position.x * cr - position.y * sr + yp * shearX) * sx;
         double y = yp * sy;
              if(x < 0 || (y > x && x < 5 - epsilon)) x += epsilon;
         else if(x > 5 || (y < x && x > 0 + epsilon)) x -= epsilon;
              if(y < 0 || (x > y && y < 6 - epsilon)) y += epsilon;
         else if(y > 6 || (x < y && y > 0 + epsilon)) y -= epsilon;
         if(x >= 0 && x <= 5 && y >= 0 && y <= 6)
         {
            int ix = Max(0, Min(4, (int)x));
            int iy = Max(0, Min(5, (int)y));
            if(iy == ix || iy == ix + 1)
            {
               int rhombus = ix + iy;
               bool top = x - ix > y - iy;
               face = -1;

               switch(rhombus)
               {
                  case 0: face = top ? 0 : 5; break;
                  case 2: face = top ? 1 : 6; break;
                  case 4: face = top ? 2 : 7; break;
                  case 6: face = top ? 3 : 8; break;
                  case 8: face = top ? 4 : 9; break;

                  case 1: face = top ? 10 : 15; break;
                  case 3: face = top ? 11 : 16; break;
                  case 5: face = top ? 12 : 17; break;
                  case 7: face = top ? 13 : 18; break;
                  case 9: face = top ? 14 : 19; break;
               }
               face++;
            }
         }
      }

      if(face)
      {
         // Degrees faceLon = facesCenterDodecahedronVertices[face].lon;
         int fy = (face-1) / 5, fx = (face-1) - 5*fy;
         double x = position.x - (2*fx + fy/2 + 1) * triWidth/2;
         double y = position.y - (yOffsets[fy] + 3 * centerToBase);
         GeoPoint dst;

         r = icosahedronToSphere({ face - 1, x, y }, dst);
         dst.lat = latGeocentricToGeodetic(dst.lat);

              if((Radians)dst.lon < -Pi - epsilon) dst.lon += 2*Pi;
         else if((Radians)dst.lon >  Pi + epsilon) dst.lon -= 2*Pi;

         result = { dst.lat, dst.lon };
      }
      return r;
   }

   // Converts geographic coordinates to coordinates on the icosahedron (forward projection)
   bool sphereToIcosahedron(const GeoPoint oc, ISEAFacePoint r)
   {
      bool result = false;
      GeoPoint c;
      int faces[2 + numIcosahedronFaces] = { -1, -1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19 };
      int i;
      double sinLat, cosLat;

      applyOrientation(oc, c);

      sinLat = sin(c.lat), cosLat = cos(c.lat);

      guessFace(c, faces);

      for(i = 0; i < 2 + numIcosahedronFaces; i++)
      {
         int face = faces[i];
         if(i < 2 || (faces[i] != faces[0] && faces[i] != faces[1]))
         {
            GeoPoint c0 = facesCenterDodecahedronVertices[face];
            double sinLat0 = sin(c0.lat), cosLat0 = cos(c0.lat);
            Degrees dLon = c.lon - c0.lon;
            double sinDLon = sin(dLon), cosDLon = cos(dLon);
            double cosLatCosDLon = cosLat * cosDLon;
            Degrees Az_earth = atan2(cosLat * sinDLon, cosLat0 * sinLat - sinLat0 * cosLatCosDLon); // Az
            Degrees AzAdjustment = faceOrientation(face);

            Az_earth += AzAdjustment;
            while(Az_earth <     0) AzAdjustment += AzMax, Az_earth += AzMax;
            while(Az_earth > AzMax) AzAdjustment -= AzMax, Az_earth -= AzMax;
            {
               double sinAz_earth = sin(Az_earth), cosAz_earth = cos(Az_earth);
               Degrees z = acos(sinLat0 * sinLat + cosLat0 * cosLatCosDLon);
               Degrees q = atan2(tang, cosAz_earth + sinAz_earth * cotTheta);
               if(z <= q + precision)
               {
                  Degrees H = acos(sinAz_earth * sinGcosSDC2VoS - cosAz_earth * cosG);
                  double area = (Radians)(Az_earth + westVertexLon + H) * R2; // A_G and A_{ABD}
                  Degrees Az = atan2(2 * area, Rprime2Tan2g - area * 2 * cotTheta); // Az'
                  double sinAz = sin(Az), cosAz = cos(Az);
                  Degrees q = atan2(tang, (cosAz_earth + sinAz_earth * cotTheta));
                  double d = RprimeTang / (cosAz + sinAz * cotTheta); // d'
                  double f = d / (Rprime2X * sin(q / 2));
                  double rho = Rprime2X * f * sin(z / 2);

                  Az -= AzAdjustment;
                  r = { face = face, x = rho * sin(Az), y = rho * cos(Az) };
                  result = true;
                  break;
               }
            }
         }
      }
      if(!result)
         r = { face = -1 };
      return result;
   }

   // Converts coordinates on the icosahedron to geographic coordinates (inverse projection)
   bool icosahedronToSphere(const ISEAFacePoint c, GeoPoint r)
   {
      if(c.face >= 0 && c.face < numIcosahedronFaces)
      {
         Degrees Az = atan2(c.x, c.y); // Az'
         double rho = sqrt(c.x * c.x + c.y * c.y);
         Degrees AzAdjustment = faceOrientation(c.face);

         Az += AzAdjustment;
         while(Az <     0) AzAdjustment += AzMax, Az += AzMax;
         while(Az > AzMax) AzAdjustment -= AzMax, Az -= AzMax;
         {
            double sinAz = sin(Az), cosAz = cos(Az);
            double cotAz = cosAz / sinAz;
            double area = Rprime2Tan2g / (2 * (cotAz + cotTheta)); // A_G or A_{ABD}
            Degrees deltaAz = 10 * precision;
            Degrees degAreaOverR2Plus180Minus36 = (Degrees) Radians { area / R2 } - westVertexLon;
            Degrees Az_earth = Az;

            while(fabs(deltaAz) > precision)
            {
               double sinAzEarth = sin(Az_earth), cosAzEarth = cos(Az_earth);
               Degrees H = acos(sinAzEarth * sinGcosSDC2VoS - cosAzEarth * cosG);
               Degrees FAz_earth = degAreaOverR2Plus180Minus36 - H - Az_earth; // F(Az) or g(Az)
               double F2Az_earth = (cosAzEarth * sinGcosSDC2VoS + sinAzEarth * cosG) / sin(H) - 1; // F'(Az) or g'(Az)
               deltaAz = -FAz_earth / F2Az_earth; // Delta Az^0 or Delta Az
               Az_earth += deltaAz;
            }
            {
               double sinAz_earth = sin(Az_earth), cosAz_earth = cos(Az_earth);
               double q = atan2(tang, (cosAz_earth + sinAz_earth * cotTheta));
               double d = RprimeTang / (cosAz + sinAz * cotTheta); // d'
               double f = d / (Rprime2X * sin(q / 2)); // f
               Degrees z = 2 * asin(rho / (Rprime2X * f));

               Az_earth -= AzAdjustment;
               {
                  Degrees lat0 = facesCenterDodecahedronVertices[c.face].lat;
                  double sinLat0 = sin(lat0), cosLat0 = cos(lat0);
                  double sinZ = sin(z), cosZ = cos(z);
                  double cosLat0SinZ = cosLat0 * sinZ;
                  Degrees lat = asin(sinLat0 * cosZ + cosLat0SinZ * cos(Az_earth));
                  Degrees lon = facesCenterDodecahedronVertices[c.face].lon + atan2(sin(Az_earth) * cosLat0SinZ, cosZ - sinLat0 * sin(lat));

                  revertOrientation({ lat, lon }, r);
               }
            }
         }
         return true;
      }
      r = { double::inf(), double::inf() };
      return false;
   }

private:
   GeoPoint orientation;
   double cosOrientationLat, sinOrientationLat;

   // Applies the selected icosahedron orientation
   static inline void applyOrientation(const GeoPoint c, GeoPoint r)
   {
      if(orientation.lat || orientation.lon)
      {
         Degrees lon = c.lon + orientation.lon;
         double sinLat = sin(c.lat), cosLat = cos(c.lat);
         double sinLon = sin(lon),  cosLon = cos(lon);
         double cosLoncosLat = cosLon * cosLat;
         r = {
            lat = asin(sinLat * cosOrientationLat + cosLoncosLat * sinOrientationLat),
            lon = atan2(sinLon * cosLat, cosLoncosLat * cosOrientationLat - sinLat * sinOrientationLat)
         };
      }
      else
         r = c;
   }

   // Inverse of applyOrientation()
   static inline void revertOrientation(const GeoPoint c, GeoPoint r)
   {
      Degrees lon = (c.lat < Degrees { -90 } + precisionPerDefinition || c.lat > Degrees { 90 } - precisionPerDefinition) ? 0 : c.lon;
      if(orientation.lat || orientation.lon)
      {
         double sinLat = sin(c.lat), cosLat = cos(c.lat);
         double sinLon = sin(lon),   cosLon = cos(lon);
         double cosLonCosLat = cosLon * cosLat;
         r = {
            asin(sinLat * cosOrientationLat - cosLonCosLat * sinOrientationLat),
            atan2(sinLon * cosLat, cosLonCosLat * cosOrientationLat + sinLat * sinOrientationLat) - orientation.lon
         };
      }
      else
         r = { c.lat, lon };
   }

   static inline void ::guessFace(const GeoPoint c, int result[2])
   {
      Degrees lat = c.lat, lon = c.lon;
      if(lat > E - F)
      {
              if (lon < Degrees { -108 }) result[0] =  0, result[1] =  5;
         else if (lon < Degrees {  -36 }) result[0] =  1, result[1] =  6;
         else if (lon < Degrees {   36 }) result[0] =  2, result[1] =  7;
         else if (lon < Degrees {  108 }) result[0] =  3, result[1] =  8;
         else                             result[0] =  4, result[1] =  9;
      }
      else if(lat < F - E)
      {
              if (lon < Degrees { -144 }) result[0] = 19, result[1] = 14;
         else if (lon < Degrees {  -72 }) result[0] = 15, result[1] = 10;
         else if (lon < Degrees {    0 }) result[0] = 16, result[1] = 11;
         else if (lon < Degrees {   72 }) result[0] = 17, result[1] = 12;
         else if (lon < Degrees {  144 }) result[0] = 18, result[1] = 13;
         else                             result[0] = 19, result[1] = 14;
      }
      else
      {
              if (lon < Degrees { -144 }) result[0] =  5, result[1] = 14;
         else if (lon < Degrees { -108 }) result[0] =  5, result[1] = 10;
         else if (lon < Degrees {  -72 }) result[0] =  6, result[1] = 10;
         else if (lon < Degrees {  -36 }) result[0] =  6, result[1] = 11;
         else if (lon < Degrees {    0 }) result[0] =  7, result[1] = 11;
         else if (lon < Degrees {   36 }) result[0] =  7, result[1] = 12;
         else if (lon < Degrees {   72 }) result[0] =  8, result[1] = 12;
         else if (lon < Degrees {  108 }) result[0] =  8, result[1] = 13;
         else if (lon < Degrees {  144 }) result[0] =  9, result[1] = 13;
         else                             result[0] =  9, result[1] = 14;
      }
   }

   static inline Degrees ::faceOrientation(int face)
   {
      return (face <= 4 || (10 <= face && face <= 14)) ? 0 : 180;
   }
}

/*
   ISEA planar space translated to place the leftmost icosahedron vertex at 0, 0,
   rotated 60 degrees clockwise,
   sheared with a 1.0 / sqrt(3) factor horizontally,
   and scaled by 1.0 / 7674457.9483758880087 m, 1.0 / 6646275.5435688359285 m,
   resulting in a 0..5 x 0..6 space with 10 square tiles where
   int(y) == int(x) || int(y) == int(x) + 1 corresponding to the diamonds
   made up of 2 faces of the icosahedron joined at their base
*/
public class ISEA2DTMS5x6Projection : ISEAPlanarProjection
{
public:
   bool geoToCartesian(const GeoPoint3D position, Vector3D result, int lonQuad)
   {
      Vector3D v;
      bool r = ISEAPlanarProjection::geoToCartesian(position, v, lonQuad);
      if(r)
      {
         static const double sr = -0.86602540378443864676372317075293618347140262690519, cr = 0.5;    // sin(-60), cos(-60)
         static const double shearX = 1.0 / 1.73205080756887729352744634150587236694280525381038; //sqrt(3); // 0.5773502691896257 -- 2*centerToBase / triWidth
         static const double sx = 1.0 / triWidth, sy = 1.0 / (3*centerToBase);
         double yp = -(v.x * sr + v.y * cr);
         result =
         {
            x = (v.x * cr - v.y * sr + yp * shearX) * sx,
            y = yp                                  * sy,
            z = v.z // Faced stored in z for now
         };
      }
      return r;
   }

   bool toISEAPlanar(const Vector3D position, Vector3D result)
   {
      bool r = false;
      static const double sr = 0.86602540378443864676372317075293618347140262690519, cr = 0.5;      // sin(60), cos(60)
      static const double shearX = -1.0 / 1.73205080756887729352744634150587236694280525381038; //sqrt(3); // -0.57735026918962573 -- -2*centerToBase / triWidth;
      static const double sx = triWidth, sy = 3*centerToBase;
      static const double epsilon = 1E-11; //1E-9; // 1E-11 fixes /dggs/ISEA3H/zones/Q0-0-D
      double x = position.x, y = position.y;
           if(x < 0 || (y > x && x < 5 - epsilon)) x += epsilon;
      else if(x > 5 || (y < x && x > 0 + epsilon)) x -= epsilon;
           if(y < 0 || (x > y && y < 6 - epsilon)) y += epsilon;
      else if(y > 6 || (x < y && y > 0 + epsilon)) y -= epsilon;

      if(x >= 0 && x <= 5 && y >= 0 && y <= 6)
         r = true;
      {
         int ix = Max(0, Min(4, (int)floor(x)));
         int iy = Max(0, Min(5, (int)floor(y)));
         if(!(iy == ix || iy == ix + 1))
            r = false;
         {
            int rhombus = ix + iy;
            bool top = x - ix > y - iy;
            int face = -1;

            switch(rhombus)
            {
               case 0: face = top ? 0 : 5; break;
               case 2: face = top ? 1 : 6; break;
               case 4: face = top ? 2 : 7; break;
               case 6: face = top ? 3 : 8; break;
               case 8: face = top ? 4 : 9; break;

               case 1: face = top ? 10 : 15; break;
               case 3: face = top ? 11 : 16; break;
               case 5: face = top ? 12 : 17; break;
               case 7: face = top ? 13 : 18; break;
               case 9: face = top ? 14 : 19; break;
            }

            {
               double ry = y * sy, rx = x * sx + ry * shearX;
               result = { rx * cr - -ry * sr, rx * sr + -ry * cr, face + 1 };
               if(face == -1)
                  r = false;
            }
         }
      }
      return r;
   }

   bool cartesianToGeo(const Vector3D position, GeoPoint3D result)
   {
      Vector3D v;
      bool r = toISEAPlanar(position, v);
      if(r)
         r = ISEAPlanarProjection::cartesianToGeo(v, result);
      return r;
   }
}
